#! /usr/bin/env python
# -*- coding: utf-8 -*-
####################
# Copyright (c) 2013, David J Garozzo. All rights reserved.
# http://www.joyunspeakable.com

# On-Screen Keyboard plugin
#	Author: David J. Garozzo
#			david@joyunspeakable.com
#	Release 1.0 - 12/03/2013 - Initial Release
#	Use at your own risk!

import os
import sys
import time
import re

# Note the "indigo" module is automatically imported and made available inside
# our global name space by the host process.

################################################################################
class Plugin(indigo.PluginBase):
	########################################
	def __init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs):
		indigo.PluginBase.__init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs)
		self.debug = False
		#self.debug = True

	def __del__(self):
		indigo.PluginBase.__del__(self)

	########################################
	def startup(self):
		self.debugLog(u"startup called")

	def shutdown(self):
		self.debugLog(u"shutdown called")
		

	########################################
	# Actions defined in MenuItems.xml:
	####################


	####################
	# This is the method that's called to build the source ActionGroup list. Note
	# that valuesDict is read-only so any changes you make to it will be discarded.
	####################
	def sourceActionGroups(self, filter="", valuesDict=None, typeId="", targetId=0):
		self.debugLog("sourceActionGroups called with filter: %s  typeId: %s  targetId: %s" % (filter, typeId, str(targetId)))
		returnList = list()
		# if valuesDict doesn't exist yet - if this is a brand new device
		# then we just create an empty dict so the rest of the logic will
		# work correctly. Many other ways to skin that particular cat. 
		if not valuesDict:
			valuesDict = {}
		for agId in indigo.actionGroups.iterkeys():
			returnList.append((str(agId),indigo.actionGroups.get(agId).name))
		return returnList


	####################
	# This is the method that's called to build the source variables list. Note
	# that valuesDict is read-only so any changes you make to it will be discarded.
	####################
	def sourceVariables(self, filter="", valuesDict=None, typeId="", targetId=0):
		self.debugLog("sourceVariables called with filter: %s  typeId: %s  targetId: %s" % (filter, typeId, str(targetId)))
		returnList = list()
		# if valuesDict doesn't exist yet - if this is a brand new device
		# then we just create an empty dict so the rest of the logic will
		# work correctly. Many other ways to skin that particular cat. 
		if not valuesDict:
			valuesDict = {}
		for variableId in indigo.variables.iterkeys():
			returnList.append((str(variableId),indigo.variables.get(variableId).name))
		return returnList


	####################
	# This is the method that's called to generate a new Variable.
	####################
	def addNewVariable(self, valuesDict=None, typeId="", targetId=0):
		self.debugLog("addNewVariable called with typeId: %s  targetId: %s" % (typeId, str(targetId)))
		# Get the device ID of the selected device
		if valuesDict and "addNewVarText" in valuesDict:
			variableName = valuesDict["addNewVarText"]
			if variableName == "":
				return
			newVar = indigo.variable.create(str(variableName), "")
			valuesDict["addNewVarText"] = ""


	####################
	# This is the method that's called to delete the selected Variable.
	####################
	def deleteVariable(self, valuesDict=None, typeId="", targetId=0):
		self.debugLog("deleteVariable called with typeId: %s  targetId: %s" % (typeId, str(targetId)))
		# Get the device ID of the selected device
		if valuesDict and "VariableMenu" in valuesDict:
			variableSelected = valuesDict["VariableMenu"]
			self.debugLog(variableSelected)
			if variableSelected == "":
				return
			newVar = indigo.variables[int(variableSelected)]
			indigo.variable.delete(newVar)


	####################
	# This is the method that's called to add a new OSK row.
	####################
	def addOSKRow(self, valuesDict=None, typeId="", targetId=0):
		self.debugLog("addOSKRow called with typeId: %s  targetId: %s" % (typeId, str(targetId)))
		if "oskRow" in valuesDict:
			# Get the data entered in for the row
			oskRow = valuesDict["oskRow"]
			if oskRow == "":
				return
			oskDataString = valuesDict.get("oskRowsData","")
			self.debugLog("adding oskRow: %s to %s" % (oskRow, oskDataString))
			# need to encode ',' to 'COMMA'
			oskRow = oskRow.replace(',', '[COMMA]')
			if oskDataString == "":
				oskDataString = oskRow
			else:
				oskDataString += "," + str(oskRow)
			valuesDict["oskRowsData"] = oskDataString
			self.debugLog("valuesDict = " + str(valuesDict))
			# Delete the selections on both dynamic lists since we don't
			# want to preserve those across dialog runs
			valuesDict["oskRow"] = ""
			if "oskRow" in valuesDict:
				del valuesDict["oskRow"]
			# return the new dict
			return valuesDict


	####################
	# This is the method that's called to insert a new OSK row.
	####################
	def insertOSKRow(self, valuesDict=None, typeId="", targetId=0):
		self.debugLog("insertOSKRow called with typeId: %s  targetId: %s" % (typeId, str(targetId)))
		if "oskRow" in valuesDict:
			# Get the data entered in for the row
			oskRow = valuesDict["oskRow"]
			if oskRow == "":
				return
			oskRow = oskRow.replace(',','[COMMA]')
			oskDataString = valuesDict.get("oskRowsData","")
			self.debugLog("adding oskRow: %s to %s" % (oskRow, oskDataString))
			# need to encode ',' to 'COMMA'
			if oskDataString == "":
				oskDataString = oskRow
			else:
				oskRows = oskDataString.split(",")
				selectedRows = valuesDict.get("oskRowsList", [])
				if len(selectedRows) > 0:
					oskRows.insert(int(selectedRows[len(selectedRows) - 1]) + 1, oskRow)
					oskDataString = ",".join(oskRows)
				#oskDataString += "," + str(oskRow)
			valuesDict["oskRowsData"] = oskDataString
			self.debugLog("valuesDict = " + str(valuesDict))
			# Delete the selections on both dynamic lists since we don't
			# want to preserve those across dialog runs
			valuesDict["oskRow"] = ""
			if "oskRow" in valuesDict:
				del valuesDict["oskRow"]
			# return the new dict
			return valuesDict



	####################
	# This is the method that's called to clear and hide the last OSK row.
	####################
	def removeOSKRow(self, valuesDict=None, typeId="", targetId=0):
		self.debugLog("removeOSKRow called with typeId: %s  targetId: %s" % (typeId, str(targetId)))
		self.debugLog("valuesDict = " + str(valuesDict))
		if "oskRowsData" in valuesDict:
			# Get the osk rows
			oskRows = valuesDict.get("oskRowsData","").split(",")
			self.debugLog("oskRows: %s" % oskRows)
			# Get the rows they've selected in the list that they want
			# to remove
			selectedRows = valuesDict.get("oskRowsList", [])
			self.debugLog("selectedRows: %s" % selectedRows)
			# Loop through the rows to be deleted list and remove them
			# reverse the list so as we delete items, the index doesn't change on us
			for row in reversed(selectedRows):
				self.debugLog("remove row: " + row)
				if oskRows[int(row)] != "--OSK1--":
					del oskRows[int(row)]
			# Set the "oskRowsData" field back to the new list which
			# has the osk row from it.
			valuesDict["oskRowsData"] = ",".join(oskRows)
			# Delete the selections on dynamic lists since we don't
			# want to preserve those across dialog runs
			if "oskRowsList" in valuesDict:
				del valuesDict["oskRowsList"]
			return valuesDict


	####################
	# This is the method that's called to get the selected OSK row and copy it into the edit text field
	####################
	def getOSKRow(self, valuesDict=None, typeId="", targetId=0):
		self.debugLog("getOSKRow called with typeId: %s  targetId: %s" % (typeId, str(targetId)))
		self.debugLog("valuesDict = " + str(valuesDict))
		if "oskRowsData" in valuesDict:
			# Get the osk rows
			oskRows = valuesDict.get("oskRowsData","").split(",")
			self.debugLog("oskRows: %s" % oskRows)
			# Get the rows they've selected in the list that they want
			# to edit
			selectedRows = valuesDict.get("oskRowsList", [])
			self.debugLog("selectedRows: %s" % selectedRows)
			# we really only want the last one
			valuesDict["oskRow"] = oskRows[int(selectedRows[len(selectedRows) - 1])]
			return valuesDict


	####################
	# This is the method that's called to get the selected OSK row and replace it with what 
	# is in the edit text field
	####################
	def replaceOSKRow(self, valuesDict=None, typeId="", targetId=0):
		self.debugLog("replaceOSKRow called with typeId: %s  targetId: %s" % (typeId, str(targetId)))
		self.debugLog("valuesDict = " + str(valuesDict))
		if "oskRowsData" in valuesDict:
			# Get the osk rows
			oskRows = valuesDict.get("oskRowsData","").split(",")
			self.debugLog("oskRows: %s" % oskRows)
			# Get the rows they've selected in the list that they want
			# to edit
			selectedRows = valuesDict.get("oskRowsList", [])
			self.debugLog("selectedRows: %s" % selectedRows)
			oskRow = valuesDict["oskRow"]
			oskRow = oskRow.replace(',','[COMMA]')
			# we really only want the last one
			oskRows[int(selectedRows[len(selectedRows) - 1])] = oskRow
			# Set the "oskRowsData" field back to the new list which
			# has the osk row from it.
			valuesDict["oskRowsData"] = ",".join(oskRows)

			return valuesDict


	####################
	# This is the method that's called to build the oskRows list. Note
	# that valuesDict is read-only so any changes you make to it will be discarded.
	####################
	def oskRowsData(self, filter="", valuesDict=None, typeId="", targetId=0):
		self.debugLog("oskRowsData called with filter: %s  typeId: %s  targetId: %s" % (filter, typeId, str(targetId)))
		returnList = list()
		# valuesDict may be empty or None if it's a brand new device
		if valuesDict and "oskRowsData" in valuesDict:
			# Get the list of osk rows
			oskRowsDataString = valuesDict["oskRowsData"]
		else:
			# set a default
			oskRowsDataString = "--OSK1--"
		self.debugLog("oskRowsDataString: " + oskRowsDataString)
		oskRowsDataList = oskRowsDataString.split(",")
		self.debugLog("oskRowsDataList: %s" % oskRowsDataList)
		rowNum = 0
		for oskRow in oskRowsDataList:
			self.debugLog("oskRow: " + oskRow)
			returnList.append((rowNum,oskRow))
			rowNum += 1
		self.debugLog("returnList: %s" % returnList)
		return returnList




	def NavigateTo( self, keyDict, keyMatrix, currentKeyArray, destKey, oskProps ):

		self.debugLog( "NavigateTo" )
		
		if destKey in keyDict:
			self.debugLog( "Destination: " + str(keyDict[destKey] ) )
		else:
			indigo.server.log( "Couldn't find destKey [" + str(destKey) + "] in OSK" )
			return None

		multiKeySpanTopPriority = oskProps["multiKeySpanTopPriority"]
		multiKeySpanBottomPriority = oskProps["multiKeySpanBottomPriority"]
		multiKeySpanLeftPriority = oskProps["multiKeySpanLeftPriority"]
		multiKeySpanRightPriority = oskProps["multiKeySpanRightPriority"]
		spiritBoardMode = oskProps["spiritBoardMode"]

		# check to see if there are multiple source keys on this keyboard

		currentKeyArrayList = list()
		
		for cKeyArray in keyDict[currentKeyArray[3]]:

			self.debugLog( str(cKeyArray) )
			
			# let's look for all source keys that are on the same screenName
			# as the screen we are on now
			if currentKeyArray[0] == cKeyArray[0]:
				currentKeyArrayList.append( cKeyArray )
		
		if( len( currentKeyArrayList ) > 1 ):
		
			# there are more than one of the current key on the source keyboard
			
			# this means a key takes up multiple keys on a keyboard.
			
			# what takes precedence coming off of the key? Is it 
			# the left-most key? The right-most key? The top-most
			# key? The bottom-most key? Or the key that came 
			# before the one that made it to this key?
			
			# for example, if the spacebar was 5 keys wide, and you came down
			# to the spacebar from the 4th key above it, when you navigate up,
			# where do you go? Do you go back to the 4th key? Or do you go
			# to the 1st key?
			
			# this is what will determine which key you will come off to
			
			for cKeyArray in currentKeyArrayList:
			
				if( multiKeySpanTopPriority ):
					if( cKeyArray[1] < currentKeyArray[1] ):
						currentKeyArray = cKeyArray

				elif( multiKeySpanBottomPriority ):
					if( cKeyArray[1] > currentKeyArray[1] ):
						currentKeyArray = cKeyArray
						
				if( multiKeySpanLeftPriority ):
					if( cKeyArray[2] < currentKeyArray[2] ):
						currentKeyArray = cKeyArray
				
				elif( multiKeySpanRightPriority ):
					if( cKeyArray[2] > currentKeyArray[2] ):
						currentKeyArray = cKeyArray


		destKeyArrayList = list()
		
		# does the destination key exist more than one time on the keyboard?
		for dKeyArray in keyDict[destKey]:

			self.debugLog( str(dKeyArray) )
			
			# let's look for all destination keys that are on the same screenName
			# as the screen we are on now
			if currentKeyArray[0] == dKeyArray[0]:
				destKeyArrayList.append( dKeyArray )
		
		
		destKeyArray = None
		if( len( destKeyArrayList ) > 1 ):		
			self.debugLog( "multi-keys match destination on this keyboard" )
			
			# the key exists multiple times on the current keyboard.
			
			# find the closest one to the current key - look for one in the same row or
			# same column.
			
			# initialize to the first one...
			destKeyArray = destKeyArrayList[0]

			for dKeyArray in destKeyArrayList:
			
				if( abs(currentKeyArray[1] - dKeyArray[1]) < abs(currentKeyArray[1] - destKeyArray[1]) ):
					destKeyArray = dKeyArray
					
				elif( abs(currentKeyArray[2] - dKeyArray[2]) < abs(currentKeyArray[2] - destKeyArray[2]) ):
					destKeyArray = dKeyArray

			
		elif( len( destKeyArrayList ) == 1 ):
			self.debugLog( "one key match on this keyboard" )
			# the key exists exactly one time on the current keyboard
			destKeyArray = destKeyArrayList[0]
			
		else: # if( destKeyArray == None ):
			self.debugLog( "no keys match on this keyboard" )
			# the key doesn't exist on the current keyboard
		
			# now we need to figure out which keyboard we need to get to, and
			# then how to get to it.
			
			# not sure it matters which keyboard we need to get to? I mean, we
			# just need to get to another keyboard. If the key is on two other
			# keyboards, does it matter which one of the two we choose? Maybe.
			# But until I find a reason to care, I'm just going to choose the
			# first one on the list.

			destKeyArray = keyDict[destKey][0]


		# array format: screenName, row, column, key

		# are both keys on the same screenName?
		if currentKeyArray[0] != destKeyArray[0]:
		
			# if they are on different screenNames, then we need to find out how to 
			# get to the right screenName first
			self.debugLog( "diff screenName" )

			# navigate to a different screen name			
			currentKeyArray = self.NavigateTo( keyDict, keyMatrix, currentKeyArray, "[" + str(destKeyArray[0]) + "]", oskProps )
			
			# now that we are on that new screen name, we our current key isn't really the key we think
			# that it is... We're on a new keyboard.
			newKeyArray = [ destKeyArray[0], currentKeyArray[1], currentKeyArray[2], keyMatrix[ destKeyArray[0] ][ currentKeyArray[1] ][ currentKeyArray[2] ] ]
						
			self.debugLog( "newKeyArray: " + str(newKeyArray))
			currentKeyArray = newKeyArray
			
						
		self.debugLog( "currentKeyArray: " + str(currentKeyArray))
		self.debugLog( "destKeyArray: " + str(destKeyArray))
		
		# are both keys on the same screenName? They should be now!
		if currentKeyArray[0] == destKeyArray[0]:

			step = 1
			# first, let's get to the correct row
			diff = currentKeyArray[1] - destKeyArray[1]
			if diff > 0:
				step = -1
				keyPress = "UPMenuId"
			else:
				step = 1
				keyPress = "DOWNMenuId"

			currentScreen = currentKeyArray[0]
			currentRow = currentKeyArray[1]
			currentColumn = currentKeyArray[2]
			currentKey = currentKeyArray[3]

			self.debugLog( "Starting Point: " )
			self.debugLog( "currentScreen: " + str(currentScreen) )
			self.debugLog( "currentRow: " + str(currentRow) )
			self.debugLog( "currentColumn: " + str(currentColumn) )
			self.debugLog( "currentKey: " + str(currentKey) )


			# go either up or down to the appropriate row

			for move in range( currentKeyArray[1], destKeyArray[1], step ):
				currentRow = currentRow + step
				
				# note, that if I travel through a key that is multiple spaces
				# wide, I may need to adjust what column I am actually in.
						
				if( multiKeySpanLeftPriority ):
					# lets check to see if keys to the left are the same 
					# as what I am on now. If so, I really need to move to
					# be left-most.
					for columnCheck in range( currentColumn, 0, -1 ):					
						if( keyMatrix[currentScreen][currentRow][currentColumn] == keyMatrix[currentScreen][currentRow][columnCheck] ):
							currentColumn = columnCheck
						else:
							break
				
				elif( multiKeySpanRightPriority ):
					# lets check to see if keys to the right are the same 
					# as what I am on now. If so, I really need to move to
					# be right-most.
					for columnCheck in range( currentColumn, len(keyMatrix[currentScreen][currentRow]), 1 ):					
						if( keyMatrix[currentScreen][currentRow][currentColumn] == keyMatrix[currentScreen][currentRow][columnCheck] ):
							currentColumn = columnCheck
						else:
							break
				
				self.debugLog( "currentRow: " + str(currentRow) )
				self.debugLog( "currentKeyRow: " + str( keyMatrix[currentScreen][currentRow] ) )
				self.debugLog( "currentKey: " + str( keyMatrix[currentScreen][currentRow][currentColumn] ) )
				if keyMatrix[currentScreen][currentRow][currentColumn] != currentKey:
					currentKey = keyMatrix[currentScreen][currentRow][currentColumn]
					self.debugLog( keyPress )
					indigo.actionGroup.execute(oskProps[keyPress])
					
				if( currentKey == destKey ):
					break

			step = 1
			# then the correct column
			diff = currentKeyArray[2] - destKeyArray[2]
			if diff > 0:
				step = -1
				keyPress = "LEFTMenuId"
			else:
				step = 1
				keyPress = "RIGHTMenuId"

			self.debugLog( "currentColumn: " + str(currentColumn) )


			# go either left or right to the appropriate column
			# note, that if I travel through a key that is multiple spaces
			# high, I many need to adjust what row I am actually in.

			for move in range( currentKeyArray[2], destKeyArray[2], step ):
				currentColumn = currentColumn + step
				
				
				if( multiKeySpanTopPriority ):
					# lets check to see if keys to the left are the same 
					# as what I am on now. If so, I really need to move to
					# be left-most.
					for rowCheck in range( currentRow, 0, -1 ):					
						if( keyMatrix[currentScreen][currentRow][currentColumn] == keyMatrix[currentScreen][rowCheck][currentColumn] ):
							currentRow = rowCheck
						else:
							break
				
				elif( multiKeySpanBottomPriority ):
					# lets check to see if keys to the right are the same 
					# as what I am on now. If so, I really need to move to
					# be right-most.
					for rowCheck in range( currentRow, len(keyMatrix[currentScreen]), 1 ):					
						if( keyMatrix[currentScreen][currentRow][currentColumn] == keyMatrix[currentScreen][rowCheck][currentColumn] ):
							currentRow = rowCheck
						else:
							break				
				
				self.debugLog( "currentColumn: " + str(currentColumn) )
				self.debugLog( "keyMatrix ?= currentKey: " + str( keyMatrix[currentScreen][currentRow][currentColumn] ) + ":" + currentKey )
				if keyMatrix[currentScreen][currentRow][currentColumn] != currentKey:
					currentKey = keyMatrix[currentScreen][currentRow][currentColumn]
					self.debugLog( keyPress )
					indigo.actionGroup.execute(oskProps[keyPress])
					
				if( currentKey == destKey ):
					break
			
			# we should double-check that we're on the key that we really want to be on???
			# maybe we traveled through a key that spans multiple rows or columns, and it
			# threw us off our route???
			
			# maybe.... future?

			if( spiritBoardMode != True ):
				keyPress = "SELECTMenuId"
				self.debugLog( keyPress )
				indigo.actionGroup.execute(oskProps[keyPress])
			else:
				time.sleep( 3 )
			
			# ok, so now we have selected the key we wanted to push. Was this key the
			# result of a single-key modifier key? For example, the SHIFT key needed
			# to be pressed to input a capital letter, but after actually pressing the
			# capital letter, the keyboard reverted back to the previous keyboard?
			# if so, we need to revert the keyboard back!
			
			# split on commas
			destKeyScreen = destKeyArray[0].split("/")

			if( len(destKeyScreen) == 2 ):
			
				# before splitting, the screen name was something like --SHIFT/OSK1--
				# now we need to get to --OSK1--
				newKeyArray = [ "--" + destKeyScreen[1], destKeyArray[1], destKeyArray[2], keyMatrix[ "--" + destKeyScreen[1] ][ destKeyArray[1] ][ destKeyArray[2] ] ]
						
				self.debugLog( "newKeyArray: " + str(newKeyArray))
				destKeyArray = newKeyArray

		return destKeyArray
			


	########################################
	# Plugin Actions object callbacks (pluginAction is an Indigo plugin action instance)
	######################
	def doOnScreenKeyboard(self, pluginAction):
		self.debugLog("doOnScreenKeyboard action called:\n" + str(pluginAction))

		device = indigo.devices[pluginAction.deviceId]
		indigo.server.log("doOnScreenKeyboard called for " + device.name)
		
		self.debugLog("pluginProps:")
		self.debugLog( str(device.pluginProps) )
		
		# Pull out the fields we need
		
		#     DOWNMenu : 102098446 (string)
		# DOWNMenu defines the ActionGroup to execute in order to make the OSK Cursor go Down
		
		#     LEFTMenu : 838501811 (string)
		# LEFTMenu defines the ActionGroup to execute in order to make the OSK Cursor go Left

		#     RIGHTMenu : 636308213 (string)
		# RIGHTMenu defines the ActionGroup to execute in order to make the OSK Cursor go Right

		#     SELECTMenu : 767122296 (string)
		# SELECTMenu defines the ActionGroup to execute in order to make the OSK Cursor Select

		#     UPMenu : 1721717996 (string)
		# UPMenu defines the ActionGroup to execute in order to make the OSK Cursor go Up

		#     VariableMenu : 1703866277 (string)
		# VariableMenu is the name of the variable that is to be printed out on the OSK
		
		#     oskRowsData : --OSK1--,DONE 0 1 2 3 4 5 6 7 8 9,[--SHIFT--] a b c d e f g h i j,[--CAPS--] k l m n o p q r s t,[--SYMBOLS--] u v w x y z [BACKSPACE],[--IPADDR--] [SPACE] [SPACE] [SPACE] [SPACE] [SPACE] [SPACE] [LEFT] [RIGHT],--SHIFT--,--CAPS--,--SYMBOLS--,--IPADDR-- (string)
		# oskRowsData defines the OSK layout
		
		# A line that contains only "--name--" stands for the name of the OSK layout.
		# A line that has spaces in it shows the keys on that row.
		# A key that contains "[--name--]" stands for the link to the OSK layout.
		# A key that contains "[SPACE]" stands for a space.
		# A key that contains "DONE" stands for the submit button.

		DOWNMenuId = ""
		LEFTMeneId = ""
		RIGHTMenuId = ""
		SELECTMenuId = ""
		UPMenuId = ""
		VariableMenuId = ""
		OSKRowsData = ""
		multiKeySpanTopPriority = False
		multiKeySpanBottomPriority = False
		multiKeySpanLeftPriority = False
		multiKeySpanRightPriority = False
		spiritBoardMode = False
		
		# starting point of cursor
		OSKStartingKey = "DONE"
		
		# ending point of cursor (to submit)
		OSKSubmitKey = "DONE"
		
		for myProps in device.pluginProps.iterkeys():
			self.debugLog(myProps)
			self.debugLog(str(device.pluginProps[myProps]))

			if myProps == "DOWNMenu":
				DOWNMenuId = int(device.pluginProps[myProps])
			elif myProps == "LEFTMenu":
				LEFTMenuId = int(device.pluginProps[myProps])
			elif myProps == "RIGHTMenu":
				RIGHTMenuId = int(device.pluginProps[myProps])
			elif myProps == "SELECTMenu":
				SELECTMenuId = int(device.pluginProps[myProps])
			elif myProps == "UPMenu":
				UPMenuId = int(device.pluginProps[myProps])
			elif myProps == "VariableMenu":
				VariableMenuId = int(device.pluginProps[myProps])
			elif myProps == "oskRowsData":
				OSKRowsData = str(device.pluginProps[myProps])
			elif myProps == "oskStartingKey":
				OSKStartingKey = str(device.pluginProps[myProps])
			elif myProps == "oskSubmitKey":
				OSKSubmitKey = str(device.pluginProps[myProps])
			elif myProps == "oskVerticalMultiKeyspanPriority":
				if( str(device.pluginProps[myProps]) == "Top" ):
					multiKeySpanTopPriority = True
					self.debugLog("multiKeySpanTopPriority")
				elif( str(device.pluginProps[myProps]) == "Bottom" ):
					multiKeySpanBottomPriority = True
					self.debugLog("multiKeySpanBottomPriority")
			elif myProps == "oskHorizontalMultiKeyspanPriority":
				if( str(device.pluginProps[myProps]) == "Left" ):
					multiKeySpanLeftPriority = True
					self.debugLog("multiKeySpanLeftPriority")
				elif( str(device.pluginProps[myProps]) == "Right" ):
					multiKeySpanRightPriority = True
					self.debugLog("multiKeySpanRightPriority")
			elif myProps == "oskSpiritBoardMode":
				spiritBoardMode = int(device.pluginProps[myProps]) == True
		
		oskProps = dict()
		oskProps["DOWNMenuId"] = DOWNMenuId
		oskProps["LEFTMenuId"] = LEFTMenuId
		oskProps["RIGHTMenuId"] = RIGHTMenuId
		oskProps["SELECTMenuId"] = SELECTMenuId
		oskProps["UPMenuId"] = UPMenuId
		oskProps["multiKeySpanTopPriority"] = multiKeySpanTopPriority
		oskProps["multiKeySpanBottomPriority"] = multiKeySpanBottomPriority
		oskProps["multiKeySpanLeftPriority"] = multiKeySpanLeftPriority
		oskProps["multiKeySpanRightPriority"] = multiKeySpanRightPriority
		oskProps["spiritBoardMode"] = spiritBoardMode
		
		
		
		# Need to identify the starting point of the cursor.
		
		
		# Need the break out the OSK into arrays that can be searched.

		#     oskRowsData : --OSK1--,DONE 0 1 2 3 4 5 6 7 8 9,[--SHIFT--] a b c d e f g h i j,[--CAPS--] k l m n o p q r s t,[--SYMBOLS--] u v w x y z [BACKSPACE],[--IPADDR--] [SPACE] [SPACE] [SPACE] [SPACE] [SPACE] [SPACE] [LEFT] [RIGHT],--SHIFT--,--CAPS--,--SYMBOLS--,--IPADDR-- (string)
		
		# split on commas
		oskList = OSKRowsData.split(",")
		
		patternOSKScreen = re.compile('--.+--')
		
		oskScreen = ""
		rowNumber = 0
		keyDict = dict()
		keyMatrix = dict()
		
		self.debugLog( "parse oskList" )
		for row in oskList:
		
			self.debugLog( "row: " + str(row) )
			
			if patternOSKScreen.match( row ) != None:
				self.debugLog( "screen found:" + str(row) )
				
				oskScreen = row
				rowNumber = 0
				
			else:
				self.debugLog( "keyRow found:" + str(row) )
			
				keyList = row.split(" ")
				
				if not( oskScreen in keyMatrix ):
					keyMatrix[oskScreen] = [ keyList ]
				else:
					keyMatrix[oskScreen].append( keyList )
				self.debugLog( "keyMatrix len:" + str(len(keyMatrix[oskScreen]) ) )
				self.debugLog( "val: " + str(keyMatrix[oskScreen][len(keyMatrix[oskScreen]) - 1]) )

				colNumber = 0
				for key in keyList:
					if not (key in keyDict):
						keyDict[key] = [ [ oskScreen, rowNumber, colNumber, key ] ]
					else:
						keyDict[key].append( [ oskScreen, rowNumber, colNumber, key ] )
										
					colNumber += 1
								
				rowNumber += 1
		
		# what if the same key is on multiple pages? might be a problem? most likely
		
		
		# Need to loop through the characters in the contents of the VariableMenu.
		variableMenuValue = str(indigo.variables[VariableMenuId].value)
		indigo.server.log("Type On-Screen:" + variableMenuValue)
		
		if OSKStartingKey in keyDict:
			currentKeyArray = keyDict[OSKStartingKey][0]
		else:
			indigo.server.log("starting key not found on OSK: [" + OSKStartingKey + "]")
		
		for c in variableMenuValue:
			self.debugLog(c)
			if c == " ":
				c = "[SPACE]"
			elif c == ",":
				c = "[COMMA]"
			self.debugLog(c)
			currentKeyArray = self.NavigateTo( keyDict, keyMatrix, currentKeyArray, c, oskProps )
			if currentKeyArray == None:
				break

		if currentKeyArray != None:
			self.debugLog( OSKSubmitKey )
			currentKeyArray = self.NavigateTo( keyDict, keyMatrix, currentKeyArray, OSKSubmitKey, oskProps )
		
